/*
             LUFA Library
     Copyright (C) Dean Camera, 2012.

  dean [at] fourwalledcubicle [dot] com
           www.lufa-lib.org
*/

/*
  Copyright 2012  Dean Camera (dean [at] fourwalledcubicle [dot] com)
  Copyright 2010  Denver Gingerich (denver [at] ossguy [dot] com)
  Copyright 2012  Frank Zhao (me [at] frank-zhao [dot] com)

  Permission to use, copy, modify, distribute, and sell this
  software and its documentation for any purpose is hereby granted
  without fee, provided that the above copyright notice appear in
  all copies and that both that the copyright notice and this
  permission notice and warranty disclaimer appear in supporting
  documentation, and that the name of the author not be used in
  advertising or publicity pertaining to distribution of the
  software without specific, written prior permission.

  The author disclaim all warranties with regard to this
  software, including all implied warranties of merchantability
  and fitness.  In no event shall the author be liable for any
  special, indirect or consequential damages or any damages
  whatsoever resulting from loss of use, data or profits, whether
  in an action of contract, negligence or other tortious action,
  arising out of or in connection with the use or performance of
  this software.
*/

/** \file
 *
 *  Main source file for the NehebkauKeyboardMouse. This file contains the main tasks of the app and
 *  is responsible for the initial application hardware configuration.
 *
 *  This device is a slave that acts as a keyboard and mouse, a master device is connected to this device
 *  using a 9 bit serial port, and transmits the keyboard and mouse HID reports through the serial port
 *
 */

#include "NehebkauKeyboardMouse.h"

/** Indicates what report mode the host has requested, true for normal HID reporting mode, false for special boot
 *  protocol reporting mode.
 */
static bool UsingReportProtocol = true;

/** Current Idle period. This is set by the host via a Set Idle HID class request to silence the device's reports
 *  for either the entire idle duration, or until the report status changes (e.g. the user presses a key).
 */
static uint16_t IdleCount = 500;

/** Current Idle period remaining. When the IdleCount value is set, this tracks the remaining number of idle
 *  milliseconds. This is separate to the IdleCount timer and is incremented and compared as the host may request
 *  the current idle period via a Get Idle HID class request, thus its value must be preserved.
 */
static uint16_t IdleMSRemaining = 0;

/** Global structure to hold the current keyboard interface HID report, for transmission to the host */
static KeyboardReport_Data_t KeyboardReportData;

/** Global structure to hold the current mouse interface HID report, for transmission to the host */
static MouseReport_Data_t MouseReportData;

// serial port related variables and definitions
#define SER_RXBUFFSIZE (HID_EPSIZE*4)
static volatile uint8_t ser_rxBuff[SER_RXBUFFSIZE];
static volatile uint8_t ser_lastHeader = 0;
static volatile uint8_t ser_hasFlags = 0;
static volatile uint8_t ser_rxBuffWritePtr = 0;

static volatile uint8_t replyToMaster = 0;

/** Main program entry point. This routine configures the hardware required by the application, then
 *  enters a loop to run the application tasks in sequence.
 */
int main(void)
{
    SetupHardware();

    sei();

    for (;;)
    {
        Keyboard_Task();
        Mouse_Task();
        USB_USBTask();
    }
}

/** Configures the board hardware and chip peripherals for the demo's functionality. */
void SetupHardware(void)
{
    /* Disable watchdog if enabled by bootloader/fuses */
    MCUSR &= ~(1 << WDRF);
    wdt_disable();

    /* Disable clock division */
    clock_prescale_set(clock_div_1);

    /* Hardware Initialization */
    USB_Init();
    Ser_Init();
    // reset global vars
    ser_rxBuffWritePtr = 0;
    ser_lastHeader = 0;
    replyToMaster = 0;
}

/** Event handler for the USB_Connect event. This indicates that the device is enumerating via the status LEDs and
 *  starts the library USB task to begin the enumeration and USB management process.
 */
void EVENT_USB_Device_Connect(void)
{
    /* Default to report protocol on connect */
    UsingReportProtocol = true;
    
    // notify the master of this event
    replyToMaster |= _BV(5);
    Ser_Tx(replyToMaster);
}

/** Event handler for the USB_Disconnect event. This indicates that the device is no longer connected to a host via
 *  the status LEDs.
 */
void EVENT_USB_Device_Disconnect(void)
{
    // notify the master of this event
    replyToMaster &= ~_BV(5);
    Ser_Tx(replyToMaster);
}

/** Event handler for the USB_ConfigurationChanged event. This is fired when the host sets the current configuration
 *  of the USB device after enumeration, and configures the keyboard device endpoints.
 */
void EVENT_USB_Device_ConfigurationChanged(void)
{
    bool ConfigSuccess = true;

    /* Setup HID Report Endpoints */
    ConfigSuccess &= Endpoint_ConfigureEndpoint(KEYBOARD_IN_EPNUM, EP_TYPE_INTERRUPT, ENDPOINT_DIR_IN,
                                                HID_EPSIZE, ENDPOINT_BANK_SINGLE);
    ConfigSuccess &= Endpoint_ConfigureEndpoint(KEYBOARD_OUT_EPNUM, EP_TYPE_INTERRUPT, ENDPOINT_DIR_OUT,
                                                HID_EPSIZE, ENDPOINT_BANK_SINGLE);
                                                
    ConfigSuccess &= Endpoint_ConfigureEndpoint(MOUSE_EPNUM, EP_TYPE_INTERRUPT, ENDPOINT_DIR_IN,
                                                HID_EPSIZE, ENDPOINT_BANK_SINGLE);

    /* Turn on Start-of-Frame events for tracking HID report period expiry */
    USB_Device_EnableSOFEvents();
}

/** Event handler for the USB_ControlRequest event. This is used to catch and process control requests sent to
 *  the device from the USB host before passing along unhandled control requests to the library for processing
 *  internally.
 */
void EVENT_USB_Device_ControlRequest(void)
{
    uint8_t* ReportData;
    uint8_t  ReportSize;

    /* Handle HID Class specific requests */
    switch (USB_ControlRequest.bRequest)
    {
        case HID_REQ_GetReport:
            if (USB_ControlRequest.bmRequestType == (REQDIR_DEVICETOHOST | REQTYPE_CLASS | REQREC_INTERFACE))
            {
                Endpoint_ClearSETUP();

                /* Determine if it is the mouse or the keyboard data that is being requested */
                if (!(USB_ControlRequest.wIndex))
                {
                    ReportData = (uint8_t*)&KeyboardReportData;
                    ReportSize = sizeof(KeyboardReportData);
                }
                else
                {
                    ReportData = (uint8_t*)&MouseReportData;
                    ReportSize = sizeof(MouseReportData);
                }

                /* Write the report data to the control endpoint */
                Endpoint_Write_Control_Stream_LE(ReportData, ReportSize);
                Endpoint_ClearOUT();

                /* Clear the report data afterwards */
                memset(ReportData, 0, ReportSize);
            }

            break;
        case HID_REQ_SetReport:
            if (USB_ControlRequest.bmRequestType == (REQDIR_HOSTTODEVICE | REQTYPE_CLASS | REQREC_INTERFACE))
            {
                Endpoint_ClearSETUP();

                /* Wait until the LED report has been sent by the host */
                while (!(Endpoint_IsOUTReceived()))
                {
                    if (USB_DeviceState == DEVICE_STATE_Unattached)
                      return;
                }

                /* Read in the LED report from the host */
                uint8_t LEDStatus = Endpoint_Read_8();

                Endpoint_ClearOUT();
                Endpoint_ClearStatusStage();

                /* Process the incoming LED report */
                ProcessLEDReport(LEDStatus);
            }

            break;
        case HID_REQ_GetProtocol:
            if (USB_ControlRequest.bmRequestType == (REQDIR_DEVICETOHOST | REQTYPE_CLASS | REQREC_INTERFACE))
            {
                Endpoint_ClearSETUP();

                /* Write the current protocol flag to the host */
                Endpoint_Write_8(UsingReportProtocol);

                Endpoint_ClearIN();
                Endpoint_ClearStatusStage();
            }

            break;
        case HID_REQ_SetProtocol:
            if (USB_ControlRequest.bmRequestType == (REQDIR_HOSTTODEVICE | REQTYPE_CLASS | REQREC_INTERFACE))
            {
                Endpoint_ClearSETUP();
                Endpoint_ClearStatusStage();

                /* Set or clear the flag depending on what the host indicates that the current Protocol should be */
                UsingReportProtocol = (USB_ControlRequest.wValue != 0);
            }

            break;
        case HID_REQ_SetIdle:
            if (USB_ControlRequest.bmRequestType == (REQDIR_HOSTTODEVICE | REQTYPE_CLASS | REQREC_INTERFACE))
            {
                Endpoint_ClearSETUP();
                Endpoint_ClearStatusStage();

                /* Get idle period in MSB, IdleCount must be multiplied by 4 to get number of milliseconds */
                IdleCount = ((USB_ControlRequest.wValue & 0xFF00) >> 6);
            }

            break;
        case HID_REQ_GetIdle:
            if (USB_ControlRequest.bmRequestType == (REQDIR_DEVICETOHOST | REQTYPE_CLASS | REQREC_INTERFACE))
            {
                Endpoint_ClearSETUP();

                /* Write the current idle duration to the host, must be divided by 4 before sent to host */
                Endpoint_Write_8(IdleCount >> 2);

                Endpoint_ClearIN();
                Endpoint_ClearStatusStage();
            }

            break;
    }
}

/** Event handler for the USB device Start Of Frame event. */
void EVENT_USB_Device_StartOfFrame(void)
{
    /* One millisecond has elapsed, decrement the idle time remaining counter if it has not already elapsed */
    if (IdleMSRemaining)
      IdleMSRemaining--;
}

/** Fills the given HID report data structure with the next HID report to send to the host.
 *
 *  \param[out] ReportData  Pointer to a HID report data structure to be filled
 */
void CreateKeyboardReport(KeyboardReport_Data_t* ReportData)
{
    /* Clear the report contents */
    memset(ReportData, 0, sizeof(KeyboardReport_Data_t));
    
    /* if we get a complete set of keyboard data from the master, then copy it to the report */  
    if ((ser_hasFlags & _BV(NEH_HEADER_KEYBOARD)) != 0)
    {
        ser_hasFlags &= ~_BV(NEH_HEADER_KEYBOARD); // clear the flag
        //ReportData->ReportID = 1;
        ReportData->Modifier = ser_rxBuff[HID_EPSIZE + 0];
        ReportData->Reserved = 0;
        ReportData->KeyCode[0] = ser_rxBuff[HID_EPSIZE + 1];
        ReportData->KeyCode[1] = ser_rxBuff[HID_EPSIZE + 2];
        ReportData->KeyCode[2] = ser_rxBuff[HID_EPSIZE + 3];
        ReportData->KeyCode[3] = ser_rxBuff[HID_EPSIZE + 4];
        ReportData->KeyCode[4] = ser_rxBuff[HID_EPSIZE + 5];
        ReportData->KeyCode[5] = ser_rxBuff[HID_EPSIZE + 6];
    }
    
    /* reset requested from master, blank out report */
    if (ser_lastHeader == NEH_HEADER_RESET)
    {
        memset(ReportData, 0, sizeof(KeyboardReport_Data_t));
    }
}

/** Fills the given HID report data structure with the next HID report to send to the host.
 *
 *  \param[out] ReportData  Pointer to a HID report data structure to be filled
 */
void CreateMouseReport(MouseReport_Data_t* ReportData)
{
    /* Clear the report contents */
    memset(ReportData, 0, sizeof(MouseReport_Data_t));
    
    /* if we get a complete set of mouse data from the master, then copy it to the report */
    if ((ser_hasFlags & _BV(NEH_HEADER_MOUSE)) != 0)
    {
        ser_hasFlags &= ~_BV(NEH_HEADER_MOUSE); // clear the flag
        //ReportData->ReportID = 2;
        ReportData->Button = ser_rxBuff[HID_EPSIZE*2 + 0];
        ReportData->X = ser_rxBuff[HID_EPSIZE*2 + 1];
        ReportData->Y = ser_rxBuff[HID_EPSIZE*2 + 2];
        ReportData->W = ser_rxBuff[HID_EPSIZE*2 + 3];
    }
    
    /* reset requested from master, blank out report */
    if (ser_lastHeader == NEH_HEADER_RESET)
    {
        memset(ReportData, 0, sizeof(MouseReport_Data_t));
    }
}

/** Processes a received LED report, and updates the board LEDs states to match.
 *
 *  \param[in] LEDReport  LED status report from the host
 */
void ProcessLEDReport(const uint8_t LEDReport)
{
    // send the LEDs back to the master
    replyToMaster &= ~(HID_KEYBOARD_LED_CAPSLOCK | HID_KEYBOARD_LED_KATANA | HID_KEYBOARD_LED_NUMLOCK | HID_KEYBOARD_LED_SCROLLLOCK);
    replyToMaster |= LEDReport & (HID_KEYBOARD_LED_CAPSLOCK | HID_KEYBOARD_LED_KATANA | HID_KEYBOARD_LED_NUMLOCK | HID_KEYBOARD_LED_SCROLLLOCK);
    Ser_Tx(replyToMaster);
}

/** Sends the next HID report to the host, via the keyboard data endpoint. */
void SendNextKeyboardReport(void)
{
    static USB_KeyboardReport_Data_t PrevKeyboardReportData;
    USB_KeyboardReport_Data_t        KeyboardReportData;
    bool                             SendReport = true;

    /* Create the next keyboard report for transmission to the host */
    CreateKeyboardReport(&KeyboardReportData);

    /* Check to see if the report data has changed - if so a report MUST be sent */
    SendReport = (memcmp(&PrevKeyboardReportData, &KeyboardReportData, sizeof(USB_KeyboardReport_Data_t)) != 0);

    /* Check if the idle period is set and has elapsed */
    if (IdleCount && (!(IdleMSRemaining)))
    {
        /* Reset the idle time remaining counter */
        IdleMSRemaining = IdleCount;

        /* Idle period is set and has elapsed, must send a report to the host */
        SendReport = true;
    }

    /* Select the Keyboard Report Endpoint */
    Endpoint_SelectEndpoint(KEYBOARD_IN_EPNUM);

    /* Check if Keyboard Endpoint Ready for Read/Write and if we should send a new report */
    if (Endpoint_IsReadWriteAllowed() && SendReport)
    {
        /* Save the current report data for later comparison to check for changes */
        PrevKeyboardReportData = KeyboardReportData;

        /* Write Keyboard Report Data */
        Endpoint_Write_Stream_LE(&KeyboardReportData, sizeof(KeyboardReportData), NULL);

        /* Finalize the stream transfer to send the last packet */
        Endpoint_ClearIN();
    }
}

void SendNextMouseReport(void)
{    
    static USB_MouseReport_Data_t PrevMouseReportData;
    USB_MouseReport_Data_t        MouseReportData;
    bool                          SendReport;

    /* Create the next mouse report for transmission to the host */
    CreateMouseReport(&MouseReportData);

    /* Check to see if the report data has changed - if so a report MUST be sent */
    SendReport = (memcmp(&PrevMouseReportData, &MouseReportData, sizeof(USB_MouseReport_Data_t)) != 0);

    /* Override the check if the Y or X values are non-zero - we want continuous movement while the joystick
     * is being held down (via continuous reports), otherwise the cursor will only move once per joystick toggle */
    if ((MouseReportData.Y != 0) || (MouseReportData.X != 0))
      SendReport = true;

    /* Check if the idle period is set and has elapsed */
    if (IdleCount && (!(IdleMSRemaining)))
    {
        /* Reset the idle time remaining counter */
        IdleMSRemaining = IdleCount;

        /* Idle period is set and has elapsed, must send a report to the host */
        SendReport = true;
    }

    /* Select the Mouse Report Endpoint */
    Endpoint_SelectEndpoint(MOUSE_EPNUM);

    /* Check if Mouse Endpoint Ready for Read/Write and if we should send a new report */
    if (Endpoint_IsReadWriteAllowed() && SendReport)
    {
        /* Save the current report data for later comparison to check for changes */
        PrevMouseReportData = MouseReportData;

        /* Write Mouse Report Data */
        Endpoint_Write_Stream_LE(&MouseReportData, sizeof(MouseReportData), NULL);

        /* Finalize the stream transfer to send the last packet */
        Endpoint_ClearIN();
    }
}

/** Reads the next LED status report from the host from the LED data endpoint, if one has been sent. */
void ReceiveNextKeyboardReport(void)
{
    /* Select the Keyboard LED Report Endpoint */
    Endpoint_SelectEndpoint(KEYBOARD_OUT_EPNUM);

    /* Check if Keyboard LED Endpoint contains a packet */
    if (Endpoint_IsOUTReceived())
    {
        /* Check to see if the packet contains data */
        if (Endpoint_IsReadWriteAllowed())
        {
            /* Read in the LED report from the host */
            uint8_t LEDReport = Endpoint_Read_8();

            /* Process the read LED report from the host */
            ProcessLEDReport(LEDReport);
        }

        /* Handshake the OUT Endpoint - clear endpoint and ready for next report */
        Endpoint_ClearOUT();
    }
}

/** Function to manage HID report generation and transmission to the host, when in report mode. */
void Keyboard_Task(void)
{
    /* Device must be connected and configured for the task to run */
    if (USB_DeviceState != DEVICE_STATE_Configured)
      return;

    /* Send the next keypress report to the host */
    SendNextKeyboardReport();

    /* Process the LED report sent from the host */
    ReceiveNextKeyboardReport();
}

/** Task to manage HID report generation and transmission to the host, when in report mode. */
void Mouse_Task(void)
{
    /* Device must be connected and configured for the task to run */
    if (USB_DeviceState != DEVICE_STATE_Configured)
      return;

    /* Send the next mouse report to the host */
    SendNextMouseReport();
}

// starts timer1 and causes a overflow interrupt in about 0.5s
void tmr_start()
{
    TCCR1B = 0x03; // start, div 64
    TCNT1 = 0; // reset
    TIMSK1 |= _BV(TOIE1); // enable interrupt
}

// 0.5 seconds passed since last serial transmission, master might have failed, so timeout and reset states
ISR(TIMER1_OVF_vect)
{
    ser_lastHeader = 0;
    ser_rxBuffWritePtr = 0;
}

ISR(USART1_RX_vect)
{
    if (bit_is_set(UCSR1B, RXB81))
    {
        // bit 8 is set, indicating that this is a header byte
        // header bytes indicate whether to set the mouse or 
        ser_rxBuffWritePtr = 0;
        ser_lastHeader = UDR1;
    }
    else
    {
        // not header, write data into the appropriate area in the temporary RX buffer
        if (ser_lastHeader == NEH_HEADER_KEYBOARD)
        {
            // write to volatile region
            ser_rxBuff[ser_rxBuffWritePtr] = UDR1;
            ser_rxBuffWritePtr++;
            
            // check if all has arrived
            if (ser_rxBuffWritePtr >= (sizeof(KeyboardReportData) - 1))
            {
                // copy to constant region
                ser_rxBuff[HID_EPSIZE + 0] = ser_rxBuff[0];
                ser_rxBuff[HID_EPSIZE + 1] = ser_rxBuff[1];
                ser_rxBuff[HID_EPSIZE + 2] = ser_rxBuff[2];
                ser_rxBuff[HID_EPSIZE + 3] = ser_rxBuff[3];
                ser_rxBuff[HID_EPSIZE + 4] = ser_rxBuff[4];
                ser_rxBuff[HID_EPSIZE + 5] = ser_rxBuff[5];
                ser_rxBuff[HID_EPSIZE + 6] = ser_rxBuff[6];
                ser_rxBuff[HID_EPSIZE + 7] = ser_rxBuff[7];
                ser_hasFlags |= _BV(NEH_HEADER_KEYBOARD); // indicate that new keyboard data is available
            }
        }
        else if (ser_lastHeader == NEH_HEADER_MOUSE)
        {
            // write to volatile region
            ser_rxBuff[HID_EPSIZE * 2 + ser_rxBuffWritePtr] = UDR1;
            ser_rxBuffWritePtr++;
            
            // check if we've received all of it
            if (ser_rxBuffWritePtr >= sizeof(MouseReportData))
            {
                // copy to constant region
                ser_rxBuff[HID_EPSIZE*3 + 0] = ser_rxBuff[HID_EPSIZE*2 + 0];
                ser_rxBuff[HID_EPSIZE*3 + 1] = ser_rxBuff[HID_EPSIZE*2 + 1];
                ser_rxBuff[HID_EPSIZE*3 + 2] = ser_rxBuff[HID_EPSIZE*2 + 2];
                ser_rxBuff[HID_EPSIZE*3 + 3] = ser_rxBuff[HID_EPSIZE*2 + 3];
                ser_rxBuff[HID_EPSIZE*3 + 4] = ser_rxBuff[HID_EPSIZE*2 + 4];
                ser_rxBuff[HID_EPSIZE*3 + 5] = ser_rxBuff[HID_EPSIZE*2 + 5];
                ser_rxBuff[HID_EPSIZE*3 + 6] = ser_rxBuff[HID_EPSIZE*2 + 6];
                ser_rxBuff[HID_EPSIZE*3 + 7] = ser_rxBuff[HID_EPSIZE*2 + 7];
                ser_hasFlags |= _BV(NEH_HEADER_MOUSE); // indicate that new mouse data is available
            }
        }
    }
    
    tmr_start();
}